/*
 * dhcpd.c
 *
 * Moreton Bay DHCP Server
 * Copyright (C) 1999 Matthew Ramsay <matthewr@moreton.com.au>
 *			Chris Trew <ctrew@moreton.com.au>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <errno.h>
#ifndef EMBED
#include <sys/ioctl.h>
#endif

#include "dhcpd.h"
#include "arpping.h"

/* debugging */
#define DEBUG				0
#define USE_PERROR		0

#define CLEAN_UP_DELAY	60
#define PID_FILE			"/var/log/dhcpd.pid"
#define VERSION			"0.8.25-2"

#define LEASE_TIME		"\x00\x0d\x2f\x00"


#ifdef CONFIG_NETtel
#define DHCPD_CONF_FILE		"/etc/config/dhcpd.conf"
#define DHCPD_IPLIST_FILE	"/etc/config/dhcpd.iplist"
#define DHCPD_LEASES_FILE	"/etc/config/dhcpd.leases"
#else
#define DHCPD_CONF_FILE		"/etc/dhcpd.conf"
#define DHCPD_IPLIST_FILE	"/etc/dhcpd.iplist"
#define DHCPD_LEASES_FILE	"/etc/dhcpd.leases"
#endif



/* route ADD/DEL -host 255.255.255.255 eth0 */
#define ADD					1
#define DEL					2


/* prototypes */
int route_add_host(int type);
int log_pid();
int serverSocket(short listen_port);
int clientSocket(short send_from_port,short send_to_port);
int getPacket(struct dhcpMessage *packet, int server_socket);
unsigned char *getOption(unsigned char *options, int option_val);
int endOption(unsigned char **optionptr);
int addOption(unsigned char *optionptr, unsigned char code,	char datalength, char *data);
int sendOffer(int client_socket, struct dhcpMessage *oldpacket);
int sendNAK(int client_socket, struct dhcpMessage *oldpacket);
int sendACK(int client_socket, struct dhcpMessage *oldpacket);
void alarmHandler();
u_int32_t findAddr(u_int8_t *chaddr, u_int32_t xid);
int search_config_file(char *filename, char *keyword, char *value);
int addLeased(u_int32_t yiaddr, u_int8_t chaddr[16]);
int test_ip(u_int32_t ipaddr);
int searchDHCP(int timeout);
int check_if_already_leased(u_int32_t yiaddr);
int addOptionMulti(unsigned char *optionptr, unsigned char code,
				char datalength, char *data,int mul);
#ifdef CONFIG_NETtel
int commitChanges();
#endif
u_int32_t freeIPAddresses(u_int32_t leased[],int num_leased);
void print_chaddr(u_int8_t *chaddr,char *title);


/* multiple client connections */
struct dhcpOfferedAddr offeredAddr[MAX_IP_ADDR];
int offer_num = 0; /* how many offers we are currently serving */

/* globals */
/* this will be configurable via command line arg -- one day */
int debug = 1; /* used to turn off debug messages */
int offered_dirty = 0; /* offered array starts clean */

unsigned char server_ipaddr[4];


int main() {
	int server_socket;
	int client_socket;
	int bytes;
	struct dhcpMessage packet;
	unsigned char *state,*hw_addr;
	unsigned char *server_id;
	int search_result;
	
	/* server ip addr */
	int fd = -1;
	struct ifreq ifr;
	struct sockaddr_in *sin;

	/* by default -- server id @FIXME */
	server_ipaddr[0] = 0x0;
	server_ipaddr[1] = 0x0;
	server_ipaddr[2] = 0x0;
	server_ipaddr[3] = 0x0;

	openlog("dhcpd", 0, 0);
	syslog(LOG_INFO, "Moreton Bay DHCP Server (v%s) started", VERSION);
	log_pid();
	
	if((fd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) >= 0) {
		strcpy(ifr.ifr_name, "eth0");
		ifr.ifr_addr.sa_family = AF_INET;
		if (ioctl(fd, SIOCGIFADDR, &ifr) == 0) {
			sin = (struct sockaddr_in *)&ifr.ifr_addr;
#if EMBED
			sin->sin_addr.s_addr = htonl(sin->sin_addr.s_addr);
#endif
#if DEBUG
			syslog(LOG_INFO, "eth0 (server_ip) = %s", inet_ntoa(sin->sin_addr));
#endif
			memcpy(server_ipaddr, &sin->sin_addr, 4);
		}
	}

#if 0
/****************************************************************************/
#ifdef CONFIG_NETtel
#define WAIT_B4_SEARCH	10
#if DEBUG
	syslog(LOG_INFO, "Server going to sleep for %d seconds", WAIT_B4_SEARCH);
#endif
	sleep(WAIT_B4_SEARCH);
#if DEBUG		
	if(debug) {
		syslog(LOG_INFO, "Searching for other DHCP Server..");
	}
#endif
	
	search_result = searchDHCP(3);
	
	if(search_result == 1) {
		syslog(LOG_ERR, "Found existing DHCP server on network -- quiting");
		exit(-1);
	} else if(search_result == -2) {
		/* this means that the NETtel couldn't check if another
		DHCP server was on the network.. most likely because there
		is *not* a dhcp server on the network and the NETtel dhcp client
		is still trying to aquire an address. we should now continue to run the
		server..
		*/
		syslog(LOG_ERR, "Failed to check for existing DHCP server -- skipping");
	}
#endif
/****************************************************************************/
#endif
	
	signal(SIGALRM, alarmHandler);
	
	while(1) { /* loop until universe collapses */
		server_socket = serverSocket(LISTEN_PORT);
		if(server_socket == -1) {
			syslog(LOG_ERR, "couldn't create server socket -- exit");
			exit(0);
		}			

		if (offered_dirty) {
			alarm(CLEAN_UP_DELAY);/* set alarm */
#if DEBUG		
			syslog(LOG_INFO, "Alarm On");
#endif
		}

		bytes = getPacket(&packet, server_socket); /* this waits for a packet - idle */
		close(server_socket);

		alarm(0); /* turn off alarm */
#if DEBUG		
		syslog(LOG_INFO, "Alarm off");
#endif
		if(bytes == -1) /* happens when the alarm times out */
			continue;
		
		offered_dirty = 1; /* Offered Array is dirty */

#ifdef CONFIG_NETtel
		/* Now check to see if the request came from this nettel.
			if so we wish to ignore its cries for an address */
		if((state = getOption(packet.options, DHCP_CLIENT_ID)) == NULL) {
			syslog(LOG_ERR, "Couldnt get option from packet -- ignoring");
			continue;
		}
		
		state++; /* move pointer up as state[0] == ARPHRD_ETHER == 0x01 */
		hw_addr = (unsigned char *) (0xf0006000);

		/*state should now be pointing to the  hardware address as will
		hw_addr which points to a memory location inside the nettel */

#if DEBUG
		syslog(LOG_INFO, "state = %02x%02x%02x%02x%02x%02x",state[0],state[1],state[2],state[3],state[4],state[5]);
		syslog(LOG_INFO, "hw1   = %02x%02x%02x%02x%02x%02x",hw_addr[0],hw_addr[1],hw_addr[2],hw_addr[3],hw_addr[4],hw_addr[5]);
		syslog(LOG_INFO, "hw2   = %02x%02x%02x%02x%02x%02x",hw_addr[6],hw_addr[7],hw_addr[8],hw_addr[9],hw_addr[10],hw_addr[11]);
#endif		
		if(memcmp(hw_addr,state,6) == 0) {
#if DEBUG
			syslog(LOG_INFO, "Not responding to NETtel client request");
#endif
			continue; /* skip everything and listen to another request */
		} else if(memcmp(hw_addr+6,state, 6) == 0) {
#if DEBUG
			syslog(LOG_INFO, "Not responding to NETtel client request");
#endif
			continue; /* skip everything and listen to another request */
		}
#if DEBUG
		else { 
			syslog(LOG_INFO, "request from outside server");
		}
#endif
#endif
	
		if((state = getOption(packet.options, DHCP_MESSAGE_TYPE)) == NULL) {
			syslog(LOG_ERR, "Couldnt get option from packet -- ignoring");
			continue;
		}
		
		/* 1. why don't we add a 'route add -host allones eth0' here as
			an experiment.
			
			2. we should lock the DHCP client from working until route_del_host()
			
			 */
		route_add_host(ADD);
		
		if((client_socket = clientSocket(LISTEN_PORT, SEND_PORT)) == -1) {
			/* @FIXME -- sould we bail.. or just return to start of while loop? */
			syslog(LOG_ERR, "couldn't create client socket -- exit");
			exit(0);
		}
			
		switch(state[0]) {
		case DHCPDISCOVER:
#if DEBUG
			if(debug) { 
				syslog(LOG_INFO,"received a DHCPDISCOVER");
			}
#endif
			if(sendOffer(client_socket, &packet) == -1) {
				syslog(LOG_ERR, "send OFFER failed");
				close(client_socket);
				continue; /* error occoured */
			}

			break;
			
		case DHCPREQUEST:
#if DEBUG
			if(debug) { 
				syslog(LOG_INFO,"received a DHCPREQUEST");
			}
#endif
			server_id = getOption(packet.options, 0x36);
			if(server_id == NULL) {
				syslog(LOG_INFO, "get option on 0x36 failed!");
				sendNAK(client_socket, &packet);
			} else {
#if DEBUG
				syslog(LOG_INFO, "%02x%02x%02x%02x", server_id[0], server_id[1],
						server_id[2], server_id[3]);
#endif
				if(memcmp(server_id, server_ipaddr, 4) == 0) {
					if (sendACK(client_socket, &packet) == -1) {
						close(client_socket);
						continue; /* error */
					}
				} else {
					sendNAK(client_socket, &packet);
				}
			}
			break;
			
		default:
#if DEBUG
			if(debug) { 
				syslog(LOG_WARNING,"Unknown DHCP Message Type (state[0] = %02x)", state[0]);
			}
#endif
		}
		
		close(client_socket);
		route_add_host(DEL);
	}

	/* never executes */
	syslog(LOG_ERR, "exit");
	closelog();
	return 0;
}


int log_pid() {
	int fd;
	pid_t pid;
	char *pidfile = PID_FILE;

	pid = getpid();
	if((fd = open(pidfile, O_WRONLY | O_CREAT, 0660)) < 0)
		return -1;
	write(fd, (void *) &pid, sizeof(pid));
	close(fd);
}


/*
 * add new option data to the options field terminating
 * with an 0xff
 */
 
int addOption(unsigned char *optionptr, unsigned char code,
				char datalength, char *data) {
	endOption(&optionptr);
	optionptr[0] = code;
	if(code == 0xff)
		return 0;
	optionptr++;
	optionptr[0] = datalength;
	optionptr++;
	memcpy(optionptr, data, datalength);
	optionptr += datalength;
	optionptr[0] = 0xff;
	
	return 0;	
}

int addOptionMulti(unsigned char *optionptr, unsigned char code,
				char datalength, char *data,int mul) {
/* Args: datalength - size of a simgle data
*/

	int i;
	
	endOption(&optionptr);	
	optionptr[0] = code;
	if(code == 0xff)
		return 0;
	optionptr++;
	optionptr[0] = (datalength * mul);
	optionptr++;
	for (i=0;i<mul;i++) {
		memcpy(optionptr, data, datalength);
		optionptr += datalength;
	}
	optionptr[0] = 0xff;
	
	return 0;	
}


/*
 * update the option pointer to point to where the 0xff is
 */
int endOption(unsigned char **optionptr) {

	unsigned char *tmpptr = *optionptr;
	
	while(tmpptr[0] != 0xff) {
		if(tmpptr[0] == 0x00)
			continue;
		tmpptr++;
		tmpptr += tmpptr[0];
		tmpptr++;
	}
	*optionptr = tmpptr;
	return 0;
}


int serverSocket(short listen_port) {
	int server_socket;
	struct sockaddr_in server;
	int n = 1;
	
	server_socket = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	
	if(server_socket == -1) {
		syslog(LOG_ERR, "couldn't create server socket");
#if USE_PERROR
		perror("socket");
#endif
		return -1;
	}

	bzero(&server, sizeof(server));
	server.sin_family = AF_INET;
	server.sin_port = htons(listen_port);
	server.sin_addr.s_addr = INADDR_ANY;

/*#if EMBED*/
	if (setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR, (char *) &n, sizeof(n)) == -1) { 
#if DEBUG
		syslog(LOG_ERR, "couldnt set option");
#endif
	} 
/*#endif*/
	
	if(bind(server_socket, (struct sockaddr *)&server, sizeof(struct sockaddr)) == -1) {
		syslog(LOG_ERR, "couldn't bind server");
#if USE_PERROR
		perror("server bind error");
#endif
		return -1;
	}
	
	return server_socket;	
}


int clientSocket(short send_from_port, short send_to_port) {
	int n = 1;
	int client_socket;
	struct sockaddr_in client;
	
	client_socket = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if(client_socket == -1) {
		syslog(LOG_ERR, "couldn't create client socket");
#if USE_PERROR
		perror("client socket");
#endif
		return -1;
	}

#if EMBED
	if (setsockopt(client_socket, SOL_SOCKET, SO_REUSEADDR, (char *) &n, sizeof(n)) == -1) { 
#if DEBUG
		syslog(LOG_ERR, "couldnt set option.");
#endif
	} 
#endif

	setsockopt(client_socket, SOL_SOCKET, SO_BROADCAST, (char *) &n, sizeof(n));
	bzero(&client, sizeof(client));
	client.sin_family = AF_INET;
	client.sin_port = htons(send_from_port);
	client.sin_addr.s_addr = INADDR_ANY;
	
	if(bind(client_socket,(struct sockaddr *)&client, sizeof(struct sockaddr))==-1)
	{
		syslog(LOG_ERR, "couldn't bind client");
#if USE_PERROR
		perror("client bind");
#endif
		return -1;
	}
	
	bzero(&client, sizeof(client));
	client.sin_family = AF_INET;
	client.sin_port = htons(send_to_port);
	client.sin_addr.s_addr = INADDR_BROADCAST;

	if(connect(client_socket, (struct sockaddr *)&client, sizeof(struct sockaddr)) == -1) {
		syslog(LOG_ERR, "client connect failed");
#if USE_PERROR
		perror("client connect");
#endif
		return -1;
	}
	
	return client_socket;	
}


int getPacket(struct dhcpMessage *packet, int server_socket) {
	char buf[sizeof(struct dhcpMessage)];
	u_int32_t magic;
	int bytes;

#if DEBUG
	syslog(LOG_INFO, "Listening for DHCP messages on network...");
#endif
	
	memset(buf, 0, sizeof(buf));
	
	bytes = read(server_socket, (char *)buf, sizeof(buf));
#if DEBUG
	syslog(LOG_INFO, "oooh, got some!");
#endif
	if(bytes < 0) {
		if (debug) { /* outputs errors on a non blocking call so it is flagged */
#if DEBUG
			syslog(LOG_INFO, "non block - couldn't read on server socket - not bad");
#endif
		}
#if USE_PERROR
		perror("read");
#endif
		return -1;
	}

	memcpy(packet, buf, sizeof(buf));
	memcpy(&magic, &packet->cookie, 4);

	if(htonl(magic) != MAGIC) {
		syslog(LOG_ERR, "corrupt DHCP message received");
		return -1;
	}
	
	return bytes;
}


unsigned char *getOption(unsigned char *options, int option_val) {
	while(options[0] != 0xff) {
		if(options[0] == 0) {
			options++;
			continue;
		}
		if(options[0] == 0xff)
			return NULL;
		if(options[0] == option_val)
			return options+2;
		options++;
		options += options[0];
		options++;
	}
	
	return NULL; /* never executed */
}


/* send a DHCP OFFER to a DHCP DISCOVER */
int sendOffer(int client_socket, struct dhcpMessage *oldpacket) {
/*
Retn: 	-1 on failure
 	0 if the ipaddress was not leased yet
	1 if the ipaddress was alreaduy leased

*/
	FILE *in;
	struct dhcpMessage packet;
	char buf[sizeof(struct dhcpMessage)];
	int bytes;
	int n,k;
	int address_used = FALSE;
	int found_one = FALSE;

	memset(&packet, 0, sizeof(packet));
	
	packet.op = BOOTREPLY;
	packet.htype = ETH_10MB;
	packet.hlen = ETH_10MB_LEN;
	packet.xid = oldpacket->xid;
	if((packet.yiaddr = findAddr(oldpacket->chaddr, oldpacket->xid)) == 0) {
		syslog(LOG_INFO, "OFFER abandoned -- no IP addresses");
		return -1;
	}
	
	memcpy(&packet.chaddr, oldpacket->chaddr, 16);
	memcpy(&packet.cookie, "\x63\x82\x53\x63", 4);
	memcpy(&packet.options, "\xff", 1);
	
	addOption(packet.options, 0x35, 0x01, "\x02");
	addOption(packet.options, 0x36, 0x04, server_ipaddr);
	/* lease time */
	addOption(packet.options, 0x33, 0x04, LEASE_TIME);
			
	memcpy(buf, &packet, sizeof(packet));
	bytes = send(client_socket, buf, sizeof(buf), 0);
	
	if(bytes == -1) {
#if DEBUG
		syslog(LOG_ERR,"couldn't error writing to client_socket");
#endif
		return -1;
	}
	
	return 0;
}


int sendNAK(int client_socket, struct dhcpMessage *oldpacket) {
	struct dhcpMessage packet;
	char buf[sizeof(struct dhcpMessage)];
	int bytes;
#if DEBUG
	if (debug) { 
		syslog(LOG_INFO,"Sending NAK..");
	}
#endif
		
	memset(&packet, 0, sizeof(packet));
	
	packet.op = BOOTREPLY;
	packet.htype = ETH_10MB;
	packet.hlen = ETH_10MB_LEN;
	packet.xid = oldpacket->xid;
	memcpy(&packet.chaddr, oldpacket->chaddr, 16);
	memcpy(&packet.cookie, "\x63\x82\x53\x63", 4);
	memcpy(&packet.options, "\xff", 1);
	/* options should look like this:
	* 0x350106 -- NAK 
	* 0x3604serverid - server id */
	addOption(packet.options, 0x35, 0x01, "\x06");
	addOption(packet.options, 0x36, 0x04, server_ipaddr);
	
	memcpy(buf, &packet, sizeof(packet));
	bytes = send(client_socket, buf, sizeof(buf), 0);
	
	if(bytes == -1) {
#if DEBUG
		syslog(LOG_ERR, "error writing to client_socket");
#endif
#if USE_PERROR
		perror("client write");
#endif
		return -1;
	}
	return 0;
}


int sendACK(int client_socket, struct dhcpMessage *oldpacket) {
	struct dhcpMessage packet;
	char buf[sizeof(struct dhcpMessage)];
	int bytes;
	int k;
	char tmp[32];
	struct in_addr inp;
	int result = FALSE;	

#if DEBUG
	if (debug) { 
		syslog(LOG_INFO,"Sending ACK");
	}
#endif
	
	memset(&packet, 0, sizeof(packet));
	
	packet.op = BOOTREPLY;
	packet.htype = ETH_10MB;
	packet.hlen = ETH_10MB_LEN;
	packet.xid = oldpacket->xid;
	packet.ciaddr = oldpacket->ciaddr;
	memcpy(&packet.chaddr, oldpacket->chaddr, 16);
	memcpy(&packet.chaddr, oldpacket->chaddr, 16);
	memcpy(&packet.cookie, "\x63\x82\x53\x63", 4);
	memcpy(&packet.options, "\xff", 1);
	/* loop thru offeredAddr to find which addr we
	 * offered this client */
	
#if DEBUG
	syslog(LOG_INFO,"Entering cycle - Number of current offers = %d",offer_num);
#endif
	for(k=0;k<offer_num;k++) { /*cycle through the offered array */
#if DEBUG
		syslog(LOG_INFO,"cycle No - %d",k);
#endif
		
		if (memcmp(offeredAddr[k].chaddr,packet.chaddr, 16) == 0) {
#if DEBUG
			if (debug) { 
				syslog(LOG_INFO,"chaddr matches what we have in our internel offer array");
			}
#endif
			packet.yiaddr = offeredAddr[k].yiaddr;
			inp.s_addr = packet.yiaddr;
#if DEBUG
			syslog(LOG_INFO,"Sending ACK for ip_addr %s", inet_ntoa(inp));
#else
			syslog(LOG_INFO,"Serving %s", inet_ntoa(inp));
#endif
		
			/* free that offer from offeredAddr */
			
			/*S. this is confising why not just 0 them and decrease the counter
			so you dont over step its bounds backwards?!
			offeredAddr[k].xid = offeredAddr[offer_num -1].xid;
			offeredAddr[k].yiaddr = offeredAddr[offer_num -1].yiaddr; */
			
			
		   /*bzero(offeredAddr[k].chaddr,16);*/
			/* this next one crashes the nettel!? */
			/*bzero(offeredAddr[k].yiaddr,sizeof(offeredAddr[k].yiaddr)); */
			
			/* lets just decrease the array */		
			offer_num--;
			result = TRUE;
			break;
		}
	}
	
	if (result != TRUE) { /*if we cant find it in the offered array somthing has gone wrong */ 
		syslog(LOG_ERR, "No match in offer array");
		return -1;
	}
	
	/* options should look like this:
	* 0x350106 -- NAK 
	* 0x3604 serverid - server id */
	addOption(packet.options, 0x35, 0x01, "\x05");
	addOption(packet.options, 0x36, 0x04, server_ipaddr);
	addOption(packet.options, 0x33, 0x04, LEASE_TIME);

	/* open up dhcpd.conf and read subnet/router/dns */
	if((search_config_file(DHCPD_CONF_FILE, "subnet", tmp)) == 1) {
		inet_aton(tmp, &inp);
		addOption(packet.options, 0x01, 0x04, (char *)&inp.s_addr);
	}
	
	if((search_config_file(DHCPD_CONF_FILE, "router", tmp)) == 1) {
		inet_aton(tmp, &inp);
		addOption(packet.options, 0x03, 0x04, (char *)&inp.s_addr);
	}
	
	if((search_config_file(DHCPD_CONF_FILE, "dns", tmp)) == 1) {
		inet_aton(tmp, &inp);
#ifdef CONFIG_NETtel 
		/* This is so the dns server is given 3 times to prevent w9x
		   machines from timing out ~ size = 12*/
		addOptionMulti(packet.options, 0x06, 0x04, (char *)&inp.s_addr,3);
#else
		addOption(packet.options, 0x06, 0x04, (char *)&inp.s_addr);
		
#endif
	}
#ifdef CONFIG_NETtel
	else if((search_config_file("/etc/config/resolv.conf", "nameserver", tmp)) == 1) {
#else
	else if((search_config_file("/etc/resolv.conf", "nameserver", tmp)) == 1) {
#endif
		inet_aton(tmp, &inp);
		addOption(packet.options, 0x06, 0x04, (char *)&inp.s_addr);
	}
	
	if((search_config_file(DHCPD_CONF_FILE, "wins", tmp)) == 1) {
		inet_aton(tmp, &inp);
		addOption(packet.options, 0x2C, 0x04, (char *)&inp.s_addr);
	}
	
	memcpy(buf, &packet, sizeof(packet));
	bytes = send(client_socket, buf, sizeof(buf), 0);
	
	if(bytes == -1) {
#if DEBUG
		syslog(LOG_ERR,"error writing to client_socket");
#endif
		return -1;
	}

	/* write new ip to lease section of config file
	 * check that we dont write a lease that is already in the
	 * lease file (ie. reusing address since MAC is the same 
	 * if it came from the lease file already dont re add it */

	if(check_if_already_leased(packet.yiaddr) == 0) {
		addLeased(packet.yiaddr, packet.chaddr);
	}

#ifdef CONFIG_NETtel
	if(commitChanges() == -1)
		return -1;
#endif
	return 0;
}


void alarmHandler() {
	/* put our offers back to 0 */
	offer_num = 0;
	offered_dirty = 0; /*Offered array is clean*/
#if DEBUG
		syslog(LOG_INFO, "INTERUPT >> resetting offer_num");
#endif

}


u_int32_t findAddr(u_int8_t *chaddr, u_int32_t xid) {
	u_int32_t yiaddr = 0;
	u_int32_t iplist[MAX_IP_ADDR];
	u_int32_t leased[MAX_IP_ADDR];
	FILE *in = NULL;
	int n = 0;
	int k,i;
	size_t items; /* return value for fread */
	int num_ip_addr;
	int num_leased;
	u_int8_t mac_addr[16];
	u_int32_t ip_addr;
	int ip_leased = FALSE; /* is the addressed leased? */
	int already_in_offered = FALSE;

		
	/* see if this chaddr is in the offered pool first */
	/* don't add this addr to offeredAddr if chaddr already in there! */
	/* win95 has a bad habbit of changing xid's halfway thru a conversation */
	for(n=0;n<offer_num;n++) {
		if (memcmp(offeredAddr[n].chaddr,chaddr,16) == 0) {
#if DEBUG			
			syslog(LOG_INFO, "already in offered array");
#endif			
			already_in_offered = TRUE;
			break;
		}
	}
	if (already_in_offered == TRUE) {
#if DEBUG
		syslog(LOG_INFO, "already offered you an address -- have it again (%x)", offeredAddr[n].yiaddr);
#endif
		return offeredAddr[n].yiaddr;
	} else {
#if DEBUG
		syslog(LOG_INFO, "Searching for address for new client...");
#endif
	}
		
	/* open up dhcpd.iplist */
	if((in = fopen(DHCPD_IPLIST_FILE, "r")) == NULL) {
		syslog(LOG_ERR, "%s not found", DHCPD_IPLIST_FILE);
		return -1;
	}
	
	/* Read in all the values into the iplist Array*/
	while(TRUE) {
		items = fread(&iplist[n++], sizeof(u_int32_t), 1, in);
		if(items < 1)
			break;
	}
	fclose(in);

	num_ip_addr = n-1;
		
#if 0
	if (debug) {
		for(n=0;n<num_ip_addr;n++) {
			syslog(LOG_INFO,"iplist[%d] = %x", n, iplist[n]);
		}
	}
#endif
	
	if((in = fopen(DHCPD_LEASES_FILE, "r")) == NULL) {
#if DEBUG
		syslog(LOG_ERR, "dhcpd.leases not found -- no leases");
#endif
	}
	
	n=0;
	/* Read in the mac - IP pair from the leases file */
	while(TRUE) {
		if(in == NULL)
			break;
		items = fread(&mac_addr, sizeof(mac_addr), 1, in);
		if(items < 1)
			break;
		items = fread(&ip_addr, sizeof(ip_addr), 1, in);
		if(items < 1)
			break;
#if DEBUG
		if (debug) { 
			syslog(LOG_INFO,"file yielded valid MAC/IP pair  - ip_addr = %x", ip_addr);
			print_chaddr(mac_addr,"MAC");
			print_chaddr(mac_addr,"CDR");
		} 
#endif
		if(memcmp(mac_addr, chaddr, sizeof(mac_addr)) == 0) {
#if DEBUG
			if (debug) { 
				syslog(LOG_INFO,"same MAC");
			}
#endif
			/* ooh! the connecting clients MAC address
			* is already in lease file!! let's offer him
			* the address he used last time */
#if DEBUG			
			syslog(LOG_INFO,"OFFER (same MAC)");
#endif			
			
			memcpy(offeredAddr[offer_num].chaddr,chaddr,16);
			offeredAddr[offer_num].yiaddr = ip_addr;
			offer_num++;
			
			return ip_addr;
		} else { 
		/* Add it to the array for later comparison between available leases 
			and the ones that are already leased
		*/
#if DEBUG
			if (debug) { 
				syslog(LOG_INFO,"added lease to array");
			}
#endif
			leased[n++] = ip_addr;
		}
	}
	if(in != NULL)
		fclose(in);
	
	syslog(LOG_INFO, "findAddr: lease check finished");
	
	num_leased = n;
#if 0
	if (debug) { 
		syslog(LOG_INFO,"num_leases = %d", num_leased);
		for(n=0;n<num_leased;n++) {
			syslog(LOG_INFO,"leased[%d] = %x", n, leased[n]);
		}
	}
#endif
		
	/* compare the leased addresses with the actual list and
	* find the first free address */ 
	/* also check the offeredAddr array so we don't offer the same
	 * addr to simultaneously connecting clients */
	for(n=0;n<num_ip_addr;n++) {
		for(k=0;k<num_leased;k++) {
			if(iplist[n] == leased[k]) {
				/* address already leased */
#if DEBUG				
				syslog(LOG_INFO,"Address %x is leased..next",iplist[n]);
#endif				
				ip_leased = TRUE;
				break;
			}
		}
		if(ip_leased == FALSE) { /* if the ip wasnt leased*/
			/* check that it is not in offered list */
			for(i=0;i<offer_num;i++) {/*cycle thru offered array */
				if(iplist[n] == offeredAddr[i].yiaddr) { /* found our ip :( */
					if (memcmp(offeredAddr[i].chaddr,chaddr,16) == 0) { /*but we own it :) */ 
#if DEBUG				
						syslog(LOG_INFO,"already in offered array.. have it again!");
#endif				
						break;
					}
						
#if DEBUG				
					syslog(LOG_INFO,"Address already offered and not cleared yet..");
#endif				
					ip_leased = TRUE;
					break;					
				}
			}
			
			if(ip_leased == FALSE) {
						
				/* Now we test to see if the address is actually currently taken 
					if it is we should also set it to leased in the leased file */
#if EMBED
				if (test_ip(iplist[n]) == 0) {
					/* it passed the test you are free to have it */
#if DEBUG				
					syslog(LOG_INFO,"Address arped and is free");
#endif				
					yiaddr = iplist[n]; /* after setting the yiaddr for testing later */
					break;/* breaks outer loop */
				}
#if DEBUG
				else {
					/* looks like that IP is taken */
					if (debug) { 
						syslog(LOG_INFO,"Free ip is not so free");
					}
				}
#endif
#else
				yiaddr = iplist[n];
#endif
				
			} else {
#if DEBUG
				syslog(LOG_INFO, "another simultaneous client has been offered this address!!");
#endif
			}
		}
		ip_leased = FALSE;
	}
	
	/*If you wanted you could compare the num ip addresses leased (num_leased) to
	  the total num ip addresses (num_ip_addr) and determine a ratio to start 
	  freeing up ip addresses -- HERE */  
	
	if(yiaddr == 0) {
		/* no free ip addresses remain */
#if DEBUG
		syslog(LOG_INFO, "No free IP addresses found in pool -- attempting to free one");
#endif

/* If the ip addresses are full then there is somthing wrong 
	It will only occour when a network card has been replaced on a machine thus
	is very rare or there are more machines than addresses which is stoopid that
	is why we should only check this when the list is full and a machine is 
	requesting an IP address.
	Also Arp.c should be checked to make sure all addresses are correct
*/
#if EMBED
		if ((yiaddr = freeIPAddresses(leased,num_leased)) == 0) { /*No IP addresses could be freed */
#if DEBUG
			syslog(LOG_ERR, "Failed to free IP addresses -- add more addresses to pool");
#endif
		} else {
#if DEBUG
			syslog(LOG_INFO, "freeable IP address found!");
#endif
		}
#else
		syslog(LOG_INFO, "not attempting a freeipaddr -- out of addresses");
#endif
	} 
	
#if DEBUG	
	syslog(LOG_INFO,"OFFER %x", yiaddr);
#endif	
	
	memcpy(offeredAddr[offer_num].chaddr,chaddr,16); /* cp chaddr to offered array */
#if DEBUG	
	syslog(LOG_INFO,"size off.chaddr =  %d", sizeof(offeredAddr[offer_num].chaddr));
#endif	
	offeredAddr[offer_num].yiaddr = yiaddr;
	offer_num++;
	
	return yiaddr;
}


/*
 * search_config_file
 *
 * This function opens up the file specified 'filename' and searches
 * through the file for 'keyword'. If 'keyword' is found any string
 * following it is stored in 'value'.. If 'value' is NULL we assume
 * the function was called simply to determing if the keyword exists
 * in the file.
 *
 * args: filename (IN) - config filename
 *	 keyword (IN) - word to search for in config file
 *	 value (OUT) - value of keyword (if value not NULL)
 *
 * retn:	-1 on error,
 *			0 if keyword not found,
 *			1 if found
 */
int search_config_file(char *filename, char *keyword, char *value) {
	FILE *in;
	int len;
	char buffer[32], w[32],	v[32];

	if ((in = fopen(filename, "r")) == NULL) {
#if DEBUG
		syslog(LOG_ERR, "%s not found",filename);
#endif
		return -1;
	}
	
	while((fgets(buffer, 32 - 1, in)) != NULL) {
		/* check if it's what we want */
		if((sscanf(buffer, "%s %s", w, v) >= 1) && (strcmp(w, keyword) == 0)) {
			/* found it :-) */
			if(value == NULL) {
				return 1;
			} else {
				strcpy(value, v);
				fclose(in);
				/* tell them we got it */
				return 1;
			}
		}
	}

	fclose(in);
	return 0;
}


int addLeased(u_int32_t yiaddr, u_int8_t *chaddr) {
	FILE *in;

#if DEBUG
	if (debug) { 
		syslog(LOG_INFO,"Writing new lease to lease file (IP = %08x)",yiaddr);
	}
	print_chaddr(chaddr,"MAC");
#endif
	if ((in = fopen(DHCPD_LEASES_FILE, "a")) == NULL) {
		syslog(LOG_ERR, "error writing to dhcpd.leases");
		return -1;
	}
	
	fwrite(chaddr, 16, 1, in);
	fwrite(&yiaddr, sizeof(u_int32_t), 1, in);
	fclose(in);
	
	return 0;		
}


#ifdef CONFIG_NETtel
int commitChanges() {
	char value[5];
	pid_t pid;
	FILE *in;
		
	/* get the pid of flatfsd */
	if ((in = fopen("/var/log/flatfsd.pid", "r")) == NULL) {
		return -1;
	}

	if(fread(&pid, sizeof(pid_t), 1, in) > 0) {
	} else {
		fclose(in);
		return -1;
	}

	fclose(in);
	
	if((kill(pid, 10)) == -1) {
		return -1;
	}
	return 0;
}
#endif

u_int32_t freeIPAddresses (u_int32_t leased[],int num_leased) {
/* Trys to free any IP addresses that may be taken up by old
   network cards by pinging all the 
	leased IP addresses starting with the ones that were leased
	 first. Hence should be the oldest.
	The first one it finds free it will reassign.

	Args: leased - the array of IP addressesthat are leased
			num_leased - size of the array

*/
		int i;
		int found;
		
		for(i=0;i<num_leased;i++) {
#if EMBED
			found = arpping(htonl(leased[i]));
#else
			found = arpping(leased[i]);
#endif
			if (found == 1) { /* if free address */ 
			/*if it timed out without a response*/
#if DEBUG		
				if(debug) {
					syslog(LOG_INFO, "Found spare address - %x",leased[i]);
				}
#endif
				return (leased[i]);
				break;
			} else if (found == 0) { /* address used */
#if DEBUG		
				syslog(LOG_INFO, "address already active - %x",leased[i]);
#endif
			} 
		}
		return 0; /* Address not found */
}


int test_ip(u_int32_t ipaddr) {
/* 
Tests a free ip to check that it is not actually leased if it is it will
add the lease to the lease file and return 1

	Retn 0 on success (not leased to another party )
		  1 on success (actually leased to another party but lease
		  		 file successfully updated) 
		  -1 on failure (error occoured)
*/		
	int found;  
	/* This is not bzerod as it is set as the default hardware address
	if it was a valid response */
	
	u_int8_t chaddr[16] = {0x00,0x00,0x00,0x00,
			   	0x00,0x00,0x00,0x00,
				0x00,0x00,0x00,0x00,
		 	        0x00,0x00,0x00,0x00};
#if DEBUG		
	if(debug) {
		syslog(LOG_INFO, "Pinging IP to see if leased");
	}
#endif

#if EMBED	
	found = arpping(htonl(ipaddr));
#else
	found = arpping(ipaddr);
#endif
	
#if DEBUG
	syslog(LOG_INFO, "arpping returned %d", found);
#endif
	
	if (found == 1) { /* if free address */ 
		/*if it timed out without a response*/
#if DEBUG		
		if(debug) {
			syslog(LOG_INFO, "IP is free to use");
		}
#endif
		return 0;
	} else if (found == 0) { /* address is used */
#if DEBUG		
		syslog(LOG_INFO, "IP address pinged - already used");
#endif
	/* now go add it to the leased file. */
		if (addLeased(ipaddr, chaddr) == -1) {
			return 1;
		}
#if DEBUG		
		else {
			syslog(LOG_INFO, "IP address leased and HW is zerod");
		}
#endif
	}
	return -1;
}


int searchDHCP(int timeout) {
/*
Desc: Looks for a dhcp server on the network

Args: client_socket - socket to send data on
      timeout - amout of time to wait for DHCP Server reply (seconds)

Retn: 	-1 on failure
 	0 if no DHCP server found
	1 if DHCP server was found
	
Warning: This function uses sleep and should not be mixed with alarm!

*/
	struct dhcpMessage packet;
	char buf[sizeof(struct dhcpMessage)];
	int bytes;
	int tmp,i = 0;
	int server_socket,client_socket;
	
	memset(&packet, 0, sizeof(packet));
	
	packet.op = BOOTREQUEST;
	packet.htype = ETH_10MB;
	packet.hlen = ETH_10MB_LEN;
	packet.hops = 0;
	packet.xid = 0xbadb00ba;  /*long - 32bit*/
	packet.secs = 0;
	packet.flags = 0;
	packet.ciaddr = 0;
	packet.yiaddr = 0;
	packet.siaddr = 0;
	packet.giaddr = 0;
	/* packet.chaddr;  //already 0'd */
	/* packet.sname;   //already 0'd */
	/* packet.file;    //already 0'd */
	packet.cookie = 0;
	/*packet.options;  //already 0'd */
	
	memcpy(&packet.cookie, "\x63\x82\x53\x63", 4);
	memcpy(&packet.options, "\xff", 1);
	
	addOption(packet.options, 0x35, 0x01,"\x1");
			
	memcpy(buf, &packet, sizeof(packet));


	client_socket = clientSocket(SEND_PORT,LISTEN_PORT);
	if(client_socket == -1)
		return -2;
	
	bytes = send(client_socket, buf, sizeof(buf), 0);
	close(client_socket);
	
	if(bytes == -1) {
#if DEBUG
		syslog(LOG_ERR,"error writing to client_socket");
#endif
		return -1;
	}

	server_socket = serverSocket(SEND_PORT);
	
	/*sets to non blocking */
	fcntl(server_socket,F_SETFL,O_NONBLOCK); 
	
	tmp = debug; /* save the value of debug */
	debug = 0; /* turn off errors produced by noblocking reads */
	while (i < timeout) { /*loop till your dizzy*/

		bytes = getPacket(&packet, server_socket);
		
		if(bytes != -1) {
			debug = tmp; /*return to orig state */
						
#if DEBUG
			if (debug) {		
				syslog(LOG_INFO, "Found Other DHCP Server");
			}
#endif
			return 1; /* Found a server*/
		}
		sleep(1);
		i++;
	}
	
	debug = tmp; /*return to orig state */
	close(server_socket);
	
#if DEBUG		
		if(debug) {
			syslog(LOG_INFO, "No DHCP Server Found");
		}
#endif

	return 0;
}


/*
 * opens up dhcpd.leases and looks for yiaddr.. returning
 * 1 if it finds it, 0 if not.
 */
int check_if_already_leased(u_int32_t yiaddr) {
	u_int32_t lease;
	FILE *in;
	int n = 0;
	u_int8_t mac_addr[16];
	u_int32_t ip_addr;
	int num_ip_addr;
	size_t items; /* return value for fread */
	
	if((in = fopen(DHCPD_LEASES_FILE, "r")) == NULL) {
#if DEBUG
		syslog(LOG_ERR, "dhcpd.leases not found -- no leases");
#endif
		return 0;
	}
	
	/* Read in the mac - IP pair from the leases file */
	while(TRUE) {
		items = fread(&mac_addr, sizeof(mac_addr), 1, in);
		if(items < 1)
			break;
		items = fread(&ip_addr, sizeof(ip_addr), 1, in);
		if(items < 1)
			break;
#if DEBUG
		if (debug) { 
			syslog(LOG_INFO,"got a valid MAC/IP pair from dhcpd.leases");
			syslog(LOG_INFO,"ip_addr taken = %x", ip_addr);
		} 
#endif
		
		/* check if yiaddr matches ip_addr */
		if(ip_addr == yiaddr) { 
			/* ip already in lease file */
			fclose(in);
			return 1;
		}
	}
	fclose(in);
	return 0;
}

void print_chaddr(u_int8_t *chaddr,char *title) {
	/* assume size = 16 */
#if DEBUG	
	syslog(LOG_INFO,"%s = %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x %02x%02x%02x%02x",
			title,chaddr[0],chaddr[1],chaddr[2],chaddr[4],
			chaddr[4],chaddr[5],chaddr[6],chaddr[7],
			chaddr[8],chaddr[9],chaddr[10],chaddr[11],
			chaddr[12],chaddr[13],chaddr[14],chaddr[15]);
#endif			
}
	
int route_add_host(int type) {
	pid_t pid;
	char *argv[16];
	int s, argc = 0;
	
	/* route add -host 255.255.255.255 eth0 */
	if((pid = vfork()) == 0) { /* child */
		argv[argc++] = "/bin/route";
		if(type == ADD)
			argv[argc++] = "add";
		else if(type == DEL)
			argv[argc++] = "del";
		argv[argc++] = "-host";
		argv[argc++] = "255.255.255.255";
		argv[argc++] = "eth0";
		argv[argc] = NULL;
		execvp("/bin/route", argv);	
		exit(0);
	} else if (pid > 0) {
		waitpid(pid, &s, 0);
	}
	
	return 0;
}
